from .node import Node
from .network import Network
from .dataset import Dataset
from .model import Model
from .cloud import *
from .localapi import *
from ._version import __version__
from .logger import *

logger.set_verbose(False)
logger.include_timestamp(False)


from tests.template_load_asia import TemplateLoadAsia
from tests.template_load_cars import TemplateLoadCars
import unittest

class TestDatasetGetters1(TemplateLoadAsia):
    def test_get_valid_state_result(self):
        self.assertAlmostEqual(self.ds.get_result(self.network.id, self.node.id, 'yes'), 0.01039, delta=0.05)

    def test_get_invalid_state_result(self):
        with self.assertRaises(ValueError):
            self.ds.get_result(self.network.id, self.node.id, 'foobar')

    def test_get_nonexistent_result(self):
        self.assertEqual(self.ds.get_result('foo', 'bar'), None)
        self.assertEqual(self.ds.get_result(self.network.id, 'bar'), None)
        self.assertEqual(self.ds.get_result('foo', self.node.id), None)

class TestDatasetGetters2(TemplateLoadCars):
    def test_get_summary_stat(self):
        self.assertAlmostEqual(self.ds.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean, 2298, delta=50)



from tests.template_load_cars import TemplateLoadCars
import unittest
import pyagena
import tests.credentials as credentials

class TestLoadAndCalculateLocal(TemplateLoadCars):

    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)
        pyagena.local_api_activate_license(credentials.agena_key)

    def test_sensitivity_local(self):
        dsMercedes = self.model.get_dataset('mercedes')
        net = self.model.get_network()
        nTotalAnnualCost = net.get_node('total_cost')

        sensConfig = self.model.create_sensitivity_config(
            network = net.id,
            targetNode = nTotalAnnualCost.id,
            sensitivityNodes = "*",
            dataSet = dsMercedes.id,
            report_settings = {"summaryStats": ["mean", "variance"]}
        )

        sensitivityReport = pyagena.local_api_sensitivity_analysis(self.model, sensConfig)

        self.assertEqual(sensitivityReport.tornadoGraphs[0].summaryStatistic, 'mean')
        self.assertAlmostEqual(sensitivityReport.tornadoGraphs[0].graphBars[0]['valueMin'], 744, delta=10)

        nCarType = net.get_node('car_type')
        nMilesPerYear = net.get_node('Miles_per_year')
        nReliability = net.get_node('Reliability')
        dsMercedes.enter_observation(network_id=net.id, node_id=nCarType.id, value="Small")
        dsMercedes.enter_observation(network_id=net.id, node_id=nMilesPerYear.id, value=1000)
        dsMercedes.enter_observation(network_id=net.id, node_id=nReliability.id, value="Low")

        sensitivityReport = pyagena.local_api_sensitivity_analysis(self.model, sensConfig)

        self.assertEqual(sensitivityReport.tornadoGraphs[0].summaryStatistic, 'mean')
        self.assertAlmostEqual(sensitivityReport.tornadoGraphs[0].graphBars[0]['valueMin'], 208, delta=10)




import logging

def set_verbose(verbose = False):
    if verbose:
        logging.getLogger().setLevel(logging.INFO)
    else:
        logging.getLogger().setLevel(logging.ERROR)

def include_timestamp(include = False):
    if include:
        logging.basicConfig(
            force=True,
            format='%(asctime)s %(message)s',
            datefmt='%Y-%m-%dT%H:%M:%S%z')
    else:
        logging.basicConfig(
            force=True,
            format='%(message)s')
        


import unittest;
import pyagena;

class TestLocalSetup(unittest.TestCase):
    # @unittest.skip("Won't work on headless os")
    def test_local_setup(self):
        pyagena.local_api_init(verbose=True)
        pyagena.local_api_get_license_summary()


    


import os;
import unittest;
import pyagena;

class TemplateLoadAsia(unittest.TestCase):
    def setUp(self):
        
        self.model = pyagena.Model.from_cmpx(os.path.join("tests", "resources", "asia.json"))
        self.ds = self.model.get_dataset(dataset_index=0)
        self.network = self.model.get_network(network_index=0)
        self.node = self.network.get_node('T')
    


import unittest;
import pyagena;

class TestCreateAsia(unittest.TestCase):
    def test_create_asia(self):
        model = pyagena.Model()
        model.change_settings(convergence=0.001)

        net = model.create_network(id = 'Car Costs_0', name = 'Car Costs', description = '.Car Cost Estimation')

        nCarType = net.create_node(id='car_type', name='Car type',type='Ranked')
        nCarType.set_distr_type('Manual')
        nCarType.states = ['Small', 'Medium', 'Large']

        nReliability = net.create_node(id='Reliability', name='Reliability',type='Ranked', states=['High', 'Medium', 'Low'])
        nReliability.set_distr_type('Manual')

        nMaintainability = net.create_node(id='Maintainability', name='Maintainability',type='Ranked')
        nMaintainability.set_variable('miles_per_year_const', 10000.0)
        nMaintainability.set_distr_type('Expression')
        nMaintainability.set_expressions(['TNormal(wmean(2.0,car_type,1.0,Reliability), 0.01, 0, 1)'])
        nMaintainability.add_parent(nCarType)
        nMaintainability.add_parent(nReliability)

        nAnnualMaintenanceCost = net.create_node(id='maintenance_cost', name='Annual maintenance cost ($)',type='ContinuousInterval', simulated=True)
        nAnnualMaintenanceCost.add_parent(nMaintainability)
        nAnnualMaintenanceCost.set_distr_type('Partitioned')
        nAnnualMaintenanceCost.set_expressions(['TNormal(100, 100, 0, 600)', 'TNormal(200.0,150.0,0.0,600.0)', 'TNormal(500.0,150.0,0.0,600.0)'], [nMaintainability.id])

        nMilesPerGallon = net.create_node(id='miles_per_gallon', name='Miles per gallon',type='ContinuousInterval', simulated=True)
        nMilesPerGallon.add_parent(nCarType)
        nMilesPerGallon.set_distr_type('Partitioned')
        nMilesPerGallon.set_expressions(['TNormal(35.0,50.0,5.0,100.0)', 'TNormal(28.0,50.0,5.0,100.0)', 'TNormal(18.0,30.0,5.0,100.0)'], [nCarType.id])

        nFuelPrice = net.create_node(id='Fuel_price', name='Fuel price $ (gallon)',type='ContinuousInterval', simulated=True)
        nFuelPrice.set_variable('fuel_price_const', 3)
        nFuelPrice.set_distr_type('Expression')
        nFuelPrice.set_expressions(['Arithmetic(fuel_price_const)'])

        nPricePerMile = net.create_node(id='price_per_mile', name='Price per mile $',type='ContinuousInterval', simulated=True)
        nPricePerMile.add_parent(nFuelPrice)
        nPricePerMile.add_parent(nMilesPerGallon)
        nPricePerMile.set_distr_type('Expression')
        nPricePerMile.set_expressions(['Arithmetic(Fuel_price/miles_per_gallon)'])

        nMilesPerYear = net.create_node(id='Miles_per_year', name='Miles per year',type='ContinuousInterval', simulated=True)
        nMilesPerYear.set_variable('miles_per_year_const', 10000.0)
        nMilesPerYear.set_distr_type('Expression')
        nMilesPerYear.set_expressions(['Arithmetic(miles_per_year_const)'])

        nAnnualFuelCost = net.create_node(id='total_fuel_cost', name='Annual fuel cost $',type='ContinuousInterval', simulated=True)
        nAnnualFuelCost.add_parent(nMilesPerYear)
        nAnnualFuelCost.add_parent(nPricePerMile)
        nAnnualFuelCost.set_distr_type('Expression')
        nAnnualFuelCost.set_expressions(['Arithmetic(price_per_mile*Miles_per_year)'])

        nTotalAnnualCost = net.create_node(id='total_cost', name='Total annual cost $',type='ContinuousInterval', simulated=True)
        nTotalAnnualCost.add_parent(nAnnualFuelCost)
        nTotalAnnualCost.add_parent(nAnnualMaintenanceCost)
        nTotalAnnualCost.set_distr_type('Expression')
        nTotalAnnualCost.set_expressions(['Arithmetic(maintenance_cost+total_fuel_cost)'])

        ds = model.get_dataset(dataset_index=0)

        pyagena.local_api_calculate(model, [ds.id])
        result = ds.get_result(net.id, nTotalAnnualCost.id)
        self.assertAlmostEqual(result.summaryStatistics.mean, 1562.0, delta=10)
        
        ds.enter_observation(net.id, nCarType.id, 'Medium')
        ds.enter_observation(net.id, nReliability.id, 'Low')
        ds.enter_observation(net.id, nFuelPrice.id, value=3, variable_name='fuel_price_const')
        ds.enter_observation(net.id, nMilesPerYear.id, value=7000, variable_name='miles_per_year_const')
        pyagena.local_api_calculate(model, [ds.id])
        result = ds.get_result(net.id, nTotalAnnualCost.id)
        self.assertAlmostEqual(result.summaryStatistics.mean, 1118.5, delta=10)



import unittest;
import pyagena;
import tests.credentials as credentials

class TestLocalActivation(unittest.TestCase):
    def test_local_activation(self):
        pyagena.local_api_activate_license(credentials.agena_key)
        mode=pyagena.local_api_get_license_summary()['Mode']
        invalid = mode in ['', 'FreeTrial', 'TimedTrial']
        self.assertFalse(invalid)

        pyagena.local_api_activate_license(credentials.agena_key)
        mode=pyagena.local_api_get_license_summary()['Mode']
        invalid = mode in ['', 'FreeTrial', 'TimedTrial']
        self.assertFalse(invalid)

        pyagena.local_api_deactivate_license()
        mode=pyagena.local_api_get_license_summary()['Mode']
        invalid = mode in ['', 'FreeTrial', 'TimedTrial']
        self.assertTrue(invalid)
        
        pyagena.local_api_deactivate_license()
        mode=pyagena.local_api_get_license_summary()['Mode']
        invalid = mode in ['', 'FreeTrial', 'TimedTrial']
        self.assertTrue(invalid)

        pyagena.local_api_activate_license(credentials.agena_key)
        mode=pyagena.local_api_get_license_summary()['Mode']
        invalid = mode in ['', 'FreeTrial', 'TimedTrial']
        self.assertFalse(invalid)

    


from .node import Node
from .network import Network
from .dataset import Dataset, dotdict
from .model import Model

from sys import platform
import os, shutil, stat
import tempfile
import json
import subprocess
import logging

def local_api_clone():
    send_command = subprocess.run(["git", "clone", "https://github.com/AgenaRisk/api.git"], capture_output=True, text=True)

    already = "already exists and is not an empty directory"
    if already in send_command.stderr:
        raise ValueError("API clone failed - destination path 'api' already exists and is not an empty directory")
    else:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)
        logging.info("The local api environment is cloned to the working directory successfully")

def local_api_compile(verbose = False):
    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))

    checkout = subprocess.run(["git", "checkout", "master"], capture_output=True, text=True)
    if verbose:
        logging.info(checkout.stdout)
        logging.info(checkout.stderr)
    
    pull = subprocess.run(["git", "pull"], capture_output=True, text=True)
    if verbose:
        logging.info(pull.stdout)
        logging.info(pull.stderr)
    get_tag = subprocess.run(["git", "describe", "--tags", "--abbrev=0"], capture_output=True, text=True)
    tag = get_tag.stdout.strip()
    tag_comm = ['git', 'checkout', tag]

    updated = subprocess.run(tag_comm, capture_output=True, text=True)
    if verbose:
        logging.info(updated.stdout)
        logging.info(updated.stderr)
    if platform == "win32":
        send_command = subprocess.run('powershell -command "mvn clean compile -DskipTests"', capture_output=True, text=True)
    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')
        send_command = subprocess.run(['mvn', 'clean', 'compile', '-DskipTests'], capture_output=True, text=True)
        
    os.chdir(cur_wd)
    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)
    if send_command.returncode == 0:
        logging.info("The local api environment is compiled with maven successfully")
    else:
        raise ValueError("maven compile has failed")


def local_api_init(verbose = False):

    if os.path.isdir("./api"):
        def rm_dir_readonly(func, path, _):
            os.chmod(path, stat.S_IWRITE)
            func(path)

        shutil.rmtree("./api/", onerror=rm_dir_readonly)

    local_api_clone()
    local_api_compile(verbose = verbose)


def _get_license_info(mvnout):
    summary = "{" + mvnout.stdout.split("{")[1].split("}")[0].strip() + "}"
    return json.loads(summary)

def local_api_activate_license(key, verbose = False):
    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))

    if platform == "win32":
        command = 'powershell -command "mvn exec:java@activate \\"-Dexec.args=`\\"--keyActivate --key ' + key + '`\\"\\""'
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')

        command = ['mvn', 'exec:java@activate', '-Dexec.args=--keyActivate --key ' + key]
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    os.chdir(cur_wd)

    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)

    already = "Product already activated"
    invalid = "Invalid license key"
    if already in send_command.stdout:
        logging.info(already)
    elif invalid in send_command.stdout:
        raise ValueError(invalid)
    else:
        license_info = _get_license_info(send_command)
        if license_info["Mode"] == "FreeTrial":
            raise ValueError("Licence key activation failed")        
        else:
            logging.info("License key activated successfully")

def local_api_deactivate_license(verbose = False):
    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))
    
    if platform == "win32":    
        command = 'powershell -command "mvn exec:java@activate \\"-Dexec.args=`\\"--keyDeactivate`\\"\\""'
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')

        command = ['mvn', 'exec:java@activate', '-Dexec.args=--keyDeactivate']
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    os.chdir(cur_wd)

    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)

    notyet = "Product not yet activated"
    if len(send_command.stderr) > 0:
        limit_reach = "The license has reached it's allowed deactivations limit"
        if limit_reach in send_command.stderr:
            raise ValueError(limit_reach)
        elif notyet in send_command.stdout:
            logging.info(notyet)
        else:
            raise ValueError("Deactivation failed")
    else:
        if notyet in send_command.stdout:
            logging.info(notyet)
        else:
            old_key = send_command.stdout.split("Key released: ")[1].split("\n")[0]
            logging.info(f"Deactivation successful - license key {old_key} is released")

def local_api_get_license_summary(verbose = False):
    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))

    if platform == "win32":
        command = 'powershell -command "mvn exec:java@activate \\"-Dexec.args=`\\"--licenseSummary`\\"\\""'
        send_command = _exec_shell_command(command, capture_output=True, text=True)
    
    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')

        command = ['mvn', 'exec:java@activate', '-Dexec.args=--licenseSummary']
        send_command = _exec_shell_command(command, capture_output=True, text=True)
    
    os.chdir(cur_wd)

    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)

    if len(send_command.stderr) > 0:
        expired = "Trial already expired"
        if expired in send_command.stderr:
            raise ValueError(expired)
        else:
            raise ValueError("Error when attempting to show license")
        
    return _get_license_info(send_command)

def local_api_show_license(verbose = False):
    license_info = local_api_get_license_summary(verbose)
    for ix, st in license_info.items():
            print(f"{ix}: {st}")
        
def local_api_calculate(model:Model, dataset_ids = None, cache_path = None, verbose = False):
    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))
    
    tempdir = tempfile.TemporaryDirectory()

    if dataset_ids is not None:
        for ds in dataset_ids:
            if ds not in model._get_datasets():
                os.chdir(cur_wd)
                raise ValueError(f"The model does not have a dataset {ds}")
            
        data_json = model._ds_to_json(dataset_ids)
    else:
        data_json = model._ds_to_json() #if dataset ids are not specified, all datasets in the model are calculated

    model_path = os.path.join(tempdir.name, data_json[0]["id"] + "_model.cmpx")
    data_path = os.path.join(tempdir.name, data_json[0]["id"] + "_dataset.json")

    model_file = model.save_to_file(model_path, strip_data=True)
    with open(data_path, "w") as outfile:
        json.dump(data_json, outfile)

    if cache_path is None:
        out_path = os.path.join(tempdir.name, data_json[0]["id"] + "_output.json")
    else:
        out_path = cache_path

    if platform == "win32":
        if cache_path is None:
            command = 'powershell -command "mvn exec:java@calculate \\"-Dexec.args=`\\"--model \'' + model_path + '\' --out \'' + out_path + '\' --data \'' + data_path + '\'`\\"\\""'
        else:
            command = 'powershell -command "mvn exec:java@calculate \\"-Dexec.args=`\\"--directoryWorking \'' + cur_wd + '\' --model \'' + model_path + '\' --out \'' + out_path + '\' --data \'' + data_path + '\' --use-cache`\\"\\""'
        send_command = _exec_shell_command(command, capture_output=True, text=True)
    
    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')

        if cache_path is None:
            command = ['mvn', 'exec:java@calculate', '-Dexec.args=--model "' + model_path + '"  --out "' + out_path + '" --data "' + data_path + '"']
        else:
            command = ['mvn', 'exec:java@calculate', '-Dexec.args=--directoryWorking "' + cur_wd + '" --model "' + model_path + '"  --out "' + out_path + '" --data "' + data_path + '" --use-cache']
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    os.chdir(cur_wd)

    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)

    if send_command.returncode != 0:
        tempdir.cleanup()
        raise ValueError("Calculation failed")
    else:
        model._import_results(out_path)
        logging.info("The calculation is completed, the dataset in the model now contains new calculation results")
        tempdir.cleanup()
    
def local_api_sensitivity_analysis(model:Model, sens_config, verbose = False):

    def _results_to_dotdict(input):
        dot_results = dotdict(input)
        for idx, tb in enumerate(dot_results.tables):
            dot_results.tables[idx] = dotdict(dot_results.tables[idx])
        for idx, cur in enumerate(dot_results.responseCurveGraphs):
            dot_results.responseCurveGraphs[idx] = dotdict(dot_results.responseCurveGraphs[idx])
        for idx, tor in enumerate(dot_results.tornadoGraphs):
            dot_results.tornadoGraphs[idx] = dotdict(dot_results.tornadoGraphs[idx])
        dot_results.sensitivityConfig = dotdict(dot_results.sensitivityConfig)

        return dot_results

    cur_wd = os.getcwd()
    os.chdir(os.path.join('.', 'api'))

    tempdir = tempfile.TemporaryDirectory()

    model_path = os.path.join(tempdir.name, "model.cmpx")
    config_path = os.path.join(tempdir.name, "sens_config.json")
    out_path = os.path.join(tempdir.name, "output.json")

    model_file = model.save_to_file(model_path)
    with open(config_path, "w") as outfile:
        json.dump(sens_config, outfile)

    if platform == "win32":
        command = 'powershell -command "mvn exec:java@sensitivity \\"-Dexec.args=`\\"--model \'' + model_path + '\' --out \'' + out_path + '\' --config \'' + config_path + '\'`\\"\\""'
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    else:
        if not (platform == "darwin" or platform == "linux" or platform == "linux2"):
            logging.warning(f'This function was not tested for platform {platform} and may not work properly')

        command = ['mvn', 'exec:java@sensitivity', '-Dexec.args=--model "' + model_path + '"  --out "' + out_path + '" --config "' + config_path + '"']
        send_command = _exec_shell_command(command, capture_output=True, text=True)

    os.chdir(cur_wd)

    if verbose:
        logging.info(send_command.stdout)
        logging.info(send_command.stderr)

    if len(send_command.stderr) > 0:
        tempdir.cleanup()
        raise ValueError("Sensitivity analysis failed")
    else:
        with open(out_path, "r") as file:
            results_string = file.read()
        sens_results = json.loads(results_string)
        sens_results = _results_to_dotdict(sens_results)
        tempdir.cleanup()
        return(sens_results)

def _exec_shell_command(*args, **kwargs):
    send_command = subprocess.run(*args, **kwargs)
    error_marker="Only Enterprise version can run in a headless environment"
    if error_marker in send_command.stdout or error_marker in send_command.stderr:
        raise RuntimeError(error_marker)
    return send_command



import logging

class Node():

    def __init__(self, id, name=None, description=None, type=None, simulated=None, states=None):
        
        # initiating a new object with the id
        self.id = str(id)
        self.parents = [] #all nodes are created with an empty list of parents
        self.variables = []

        # assigning name and description
        if name is None:
             self.name = id
        else:
             self.name = str(name)

        if description is None:
             self.description = "New Node"
        else:
             self.description = str(description)
        
        # first check if the node is simulated or not, then get type and states information, use sensible defaults when not given
        if simulated is None or simulated==False:
            self.simulated = False
            if type is None:
                self.type = "Boolean"
                if states is None:
                    self.states = ["False", "True"]
                else:
                    if len(states)==2:
                        self.states = states
                    else:
                        self.states = ["False", "True"]
            else:
                self.type = type
                if self.type == "Labelled":
                    if states is None:
                        self.states = ["False", "True"]
                    else:
                        self.states = states
                if self.type == "Ranked":
                    if states is None:
                        self.states = ["Low", "Medium", "High"]
                    else:
                        self.states = states
                if self.type == "DiscreteReal":
                    if states is None:
                        self.states = ["0.0", "1.0"]
                    else:
                        self.states = states
                if self.type == "ContinuousInterval":
                    if states is None:
                        self.states = ["(-Infinity, -1)", "[-1, 1)", "[1, Infinity)"]
                    else:
                        self.states = states
                if self.type == "IntegerInterval":
                    if states is None:
                        self.states = ["(-Infinity, -1]", "[0, 4]"," [5, Infinity)"]
                    else:
                        self.states = states                
                if self.type == "Boolean":
                    if states is None:
                        self.states = ["False", "True"]
                    else:
                        if len(states)==2:
                            self.states = states
                        else:
                            self.states = ["False", "True"]
        else:
            self.simulated = True
            self.states = None
            if type is None:
                self.type = "ContinuousInterval" 
            else:             
                if type == "IntegerInterval":
                    self.type = type
                else:
                    self.type = "ContinuousInterval"    

        # create distribution type, probabilities, and expressions fields and fill in with sensible defaults
        # these fields will have class methods to rewrite
        if self.simulated:
            self.distr_type = "Expression"
            self.expressions = "Normal(0,1000000)"
            self.probabilities = None
        else:
            if (self.type == "ContinuousInterval" or self.type == "IntegerInterval"):
                self.distr_type = "Expression"
                self.expressions = "Normal(0,1000000)"
                self.probabilities = None
            else:
                self.distr_type = "Manual"
                self.probabilities = [[1/len(self.states)] * len(self.states)]
                self.expressions = None
        
    @property
    def type(self):
        return self.__type
    
    @type.setter
    def type(self, value):
        if value is not None and value not in ["Boolean", "Labelled", "Ranked", "DiscreteReal", "ContinuousInterval", "IntegerInterval"]:
            raise ValueError("Invalid node type")
        
        self.__type = value

    def __str__(self) -> str:
        if self.states is not None:
            return  "Node id: % s\nNode name: % s\nNode type: % s\nNode states: % s" % (self.id, self.name, self.type, ", ".join(self.states))
        else:
            return  "Node id: % s\nNode name: % s\nNode type: % s" % (self.id, self.name, self.type)

    def __repr__(self) -> str:
        if self.simulated:
            return "Simulated node: % s (% s)" % (self.name, self.type) 
        else:
            return "Node: % s (% s)" % (self.name, self.type)
    

    def set_states(self, states, from_cmpx=False):
        
        previous_states = len(self.states)

        self.states = states
        if len(self.states) != previous_states:
            self._reset_probabilities()
            if not from_cmpx:
                logging.info("The node states are updated, the NPT values are reset to uniform because the number of states has changed")
        else:
            if not from_cmpx:
                logging.info("The node states are updated")
            
    def add_parent(self, new_parent: "Node", from_cmpx=False):

        if new_parent.id == self.id:
            raise ValueError("You cannot add a node as a parent to itself")

        parent_ids = []
        if len(self.parents)>0:
            for pr in self.parents:
                parent_ids.append(pr.id)
        
        if new_parent.id in parent_ids:
            raise ValueError("The parent already exists")
        else:
            self.parents.append(new_parent)
        
        if self.distr_type == "Manual":
            if new_parent.distr_type == "Manual":
                self._reset_probabilities()
                if not from_cmpx:
                    logging.info(f"The node {new_parent.name} has been added to the parents of {self.name} and NPT values are reset to uniform")
            else:
                self.set_distr_type("Expression", from_cmpx=from_cmpx)
                if not from_cmpx:
                    logging.info(f"The node {new_parent.name} has been added to the parents of {self.name} and distribution (table) type of {self.name} is changed to expression")
        else:
            if not from_cmpx:
                logging.info(f"The node {new_parent.name} has been added to the parents of {self.name} and now can be used in its expression")

    def _addparentbyID(self, parentslist: list, parentid: str):
        par = [pr for pr in parentslist if pr.id==parentid].pop()
        self.add_parent(par, from_cmpx=True)

    def remove_parent(self, old_parent: "Node"):
        if old_parent in self.parents:
            self.parents.remove(old_parent)

        if self.distr_type == "Manual":
            self._reset_probabilities()
            logging.info(f"The node {old_parent.name} has been removed from the parents of {self.name} and NPT values are reset to uniform")
        else:
            self.expressions = "Normal(0,1000000)"
            logging.info(f"The node {old_parent.name} has been removed from the parents of {self.name} and the expression has been reset to default")

    # A function to reset the manual NPTs to default uniform when a parent is added or removed
    def _reset_probabilities(self):
        temp_length = 1
        if len(self.parents)>0:
            for pr in self.parents:
                temp_length *= len(pr.states)
        
        self.probabilities = [[1/len(self.states)] * len(self.states)] * temp_length

    def set_probabilities(self, probabilities, by_row=False, from_cmpx = False):
        
        if not by_row:
            if not self.simulated:
                if (self.distr_type == "Manual") or (self.type=="ContinuousInterval" and self.states is not None) or (self.type=="IntegerInterval" and self.states is not None):
                    if not from_cmpx:
                        temp_length = 1
                        subset_length_control = 1
                        if len(self.parents)>0:
                            for pr in self.parents:
                                if pr.states is not None:
                                    temp_length *= len(pr.states)
                    
                        for ss in probabilities:
                            if len(ss)==len(self.states):
                                subset_length_control *= 1
                            else:
                                subset_length_control *= 0
                    
                        if (len(probabilities) == temp_length) & (subset_length_control==1):
                            self.probabilities = probabilities
                        else:
                            raise ValueError("The number of probabilities does not match the size of node NPT")
                    else:
                        self.probabilities = probabilities

            else:
                raise ValueError(f"The node {self.id} does not have a manual NPT")

        if by_row:
            if not self.simulated:
                if (self.distr_type == "Manual") or (self.type=="ContinuousInterval" and self.states is not None) or (self.type=="IntegerInterval" and self.states is not None):
                    if not from_cmpx:
                        temp_length = 1
                        subset_length_control = 1
                        if len(self.parents)>0:
                            for pr in self.parents:
                                if pr.states is not None:
                                    temp_length *= len(pr.states)

                        if (self.type != "ContinuousInterval") or (self.type != "IntegerInterval"):
                            for ss in probabilities:
                                if len(ss)==temp_length:
                                    subset_length_control *= 1
                                else:
                                    subset_length_control *= 0
                    
                        if (len(probabilities) == len(self.states)) & (subset_length_control==1):
                            self.probabilities = list(map(list, zip(*probabilities)))
                        else:
                            raise ValueError("The number of probabilities does not match the size of node NPT")
                    else:
                        self.probabilities = list(map(list, zip(*probabilities)))
            else:
                raise ValueError(f"The node {self.id} does not have a manual NPT")

    def _get_parents(self):
        par_list = []
        if len(self.parents)>0:
            for pr in self.parents:
                par_list.append(pr.id)
        return par_list

    def set_expressions(self, expressions, partitioned_parents=None, from_cmpx=False):
        self.expressions = expressions
        
        if (partitioned_parents is None) & (self.distr_type == "Manual"):
            self.set_distr_type("Expression", from_cmpx=from_cmpx)

        if partitioned_parents is not None:
            if set(partitioned_parents).issubset(set(self._get_parents())):
                self.partitions = partitioned_parents
                self.set_distr_type("Partitioned", from_cmpx=from_cmpx)
            else:
                self.expressions = None
                raise ValueError(f"One or more given partition parents is not a parent of the node {self.name}")

    def _get_variable_names(self):
        var_names = []
        if len(self.variables)>0:
            for vr in self.variables:
                (k, v), = vr.items()
                var_names.append(k)
        
        return var_names
    
    def _get_variable_value(self, variable_name):
        if len(self.variables)>0:
            for vr in self.variables:
                (k, v), = vr.items()
                if k == variable_name:
                    variable_value = v
        
        return variable_value

    def set_variable(self, variable_name, variable_value, from_cmpx=False):
        if variable_name in self._get_variable_names():
            raise ValueError("The node already has a variable (constant) with this name")
        else:
            self.variables.append({variable_name:variable_value})
            if not from_cmpx:
                logging.info("The variable (constant) is successfully added to the node")

    def remove_variable(self, variable_name):
        if variable_name in self._get_variable_names():
            for ix, vr in enumerate(self.variables):
                (k, v), = vr.items()
                if k == variable_name:
                    self.variables.pop(ix)
            logging.info("The variable (constant) is removed from the node")
        else:
            raise ValueError(f"The node does not have a variable called {variable_name}")

    def set_distr_type(self, distr_type, from_cmpx=False):
        self.distr_type = distr_type

        if self.distr_type=="Manual":
            self.partitions = None
            self.expressions = None
            self._reset_probabilities()
            if not from_cmpx:
                logging.info("The node's distribution type is converted to a manual table and a default NPT is assigned")
        
        if (self.distr_type=="Expression") | (self.distr_type=="Partitioned"):
            self.probabilities = None
            if self.expressions == None:
                self.expressions = "Normal(0,1000000)"
                if not from_cmpx:
                    logging.info("The node's distribution type is converted to expression (or partitioned) and a default expression is assigned")
            else:
                if not from_cmpx:
                    logging.info("The node's distribution type is converted to expression (or partitioned)")


    @property
    def distr_type(self):
        return self.__distr_type
    
    @distr_type.setter
    def distr_type(self, value):
        if value is not None and value not in ["Manual", "Expression", "Partitioned"]:
            raise ValueError("Invalid distribution (table) type - it should be Manual, Expression, or Partitioned")
        
        if (value=="Partitioned") & (len(self.parents)==0):
            raise ValueError("The node has no parents, it cannot have a partitioned expression")
        
        if (value=="Manual") & (self.simulated):
            raise ValueError("Simulation nodes cannot have a manual table definition")

        self.__distr_type = value


    @property
    def type(self):
        return self.__type
    
    @type.setter
    def type(self, value):
        if value is not None and value not in ["Boolean", "Labelled", "Ranked", "DiscreteReal", "ContinuousInterval", "IntegerInterval"]:
            raise ValueError("Invalid node type")
        
        self.__type = value


import logging

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    
class Dataset():
    def __init__(self, id, observations = None, results=None):
         
        self.id = str(id)
        if observations is not None:
            self.observations = observations
        else:
            self.observations = []
        
        if results is not None:
            self.results = results
        else:
            self.results = []
        
        self._convert_to_dotdict()
    
    def get_result(self, network_id, node_id, node_state = None):
            result = [res for res in self.results if res.node == node_id and res.network == network_id]
            if len(result) == 0:
                return None
            else:
                result = result.pop()
                if node_state is not None:
                    all_states = [st["label"] for st in result["resultValues"]]
                    if node_state not in all_states:
                        raise ValueError(f"The result does not have a value for the state {node_state}")
                    else:
                        state_value = [rv for rv in result['resultValues'] if rv['label'] == node_state].pop()['value']
                    return state_value
                else:
                    return result

        
    def enter_observation(self, network_id, node_id, value, variable_name=None):
          
        new_obs = {"node":node_id, "network":network_id, "entries":[]}

        if variable_name is not None:
            new_obs["constantName"] = variable_name
            new_obs["entries"].append({"weight":1, "value":value})
        else:
            if isinstance(value, list) and len(value)>1:
                for vl in value:
                    new_obs["entries"].append({"weight":vl[0], "value":vl[1]})
            else:
                new_obs["entries"].append({"weight":1, "value":value})

        if self.observations is None:
            self.observations = []
          
        if len(self.observations)>0:
            obs_rewrite = False
            var_obs_check = False

            if variable_name is None:     
                for idx, obs in enumerate(self.observations):
                    if (obs["node"] == node_id) & (obs["network"] == network_id):
                        if "constantName" not in obs.keys():
                            obs_rewrite = True
                            rewrite_idx = idx
                        else:
                            obs_rewrite = False
                            var_obs_check = True
            else:
                for idx, obs in enumerate(self.observations):
                    if (obs["node"] == node_id) & (obs["network"] == network_id):
                        if "constantName" not in obs.keys():
                            obs_rewrite = False
                        else:
                            if obs["constantName"] == variable_name:
                                obs_rewrite = True
                                rewrite_idx = idx

            if obs_rewrite:
                self.observations[rewrite_idx] = new_obs
                if variable_name is None:
                    logging.info(f"The observation of {value} is entered to the node {node_id}")
                else:
                    logging.info(f"The observation of {variable_name} = {value} is entered to the node {node_id}")
            if not obs_rewrite:
                self.observations.append(new_obs)
                if var_obs_check:
                    logging.warning(f"The observation of {value} is entered to the node {node_id} with existing variable observations, in calculations the variable observations will be ignored")
                else:
                    if variable_name is None:
                        logging.info(f"The observation of {value} is entered to the node {node_id}")
                    else:
                        logging.info(f"The observation of {variable_name} = {value} is entered to the node {node_id}")

        else:
            self.observations.append(new_obs)
            if variable_name is None:
                logging.info(f"The observation of {value} is entered to the node {node_id}")
            else:
                logging.info(f"The observation of {variable_name} = {value} is entered to the node {node_id}")

    def remove_observation(self, network_id, node_id):

        obs_del = [obs for obs in self.observations if obs["node"]==node_id and obs["network"]==network_id].pop()
        self.observations.remove(obs_del)
        logging.info("The observation is successfully removed")
    
    def clear_all_observations(self):
          self.observations = []
          logging.info("All observations in the dataset are successfully cleared")

    def _convert_to_dotdict(self):
        dot_obs = []

        for ix, ob in enumerate(self.observations):
            dot_obs.append(dotdict(ob))
            for idx, ent in enumerate(ob["entries"]):
                dot_obs[ix].entries[idx] = dotdict(ent)
        
        self.observations = dot_obs

        dot_res = []

        for ix, res in enumerate(self.results):
            dot_res.append(dotdict(res))
            if "summaryStatistics" in dot_res[ix].keys():
                dot_res[ix].summaryStatistics = dotdict(res["summaryStatistics"])
            for dx, rv in enumerate(res["resultValues"]):
                dot_res[ix].resultValues[dx] = dotdict(rv)
        
        self.results = dot_res

    def __str__(self) -> str:
        if self.results is not None:
            if self.observations is not None:
                return  "Dataset id: % s\nNumber of observations: % d\nDataset contains calculation results" % (self.id, len(self.observations))
            else:
                return "Dataset id: % s\nNumber of observations: 0\nDataset contains calculation results" % (self.id)
        else:
            if self.observations is not None:
                return  "Network id: % s\nNumber of observations: % d\nDataset does not contain calculation results" % (self.id, len(self.observations))   
            else:
                return "Dataset id: % s\nNumber of observations: 0\nDataset does not contain calculation results" % (self.id)

    def __repr__(self) -> str:
        if self.observations is not None:
            return "Dataset: % s (with % d observations)" % (self.id, len(self.observations)) 
        else:
            return "Dataset: % s (with 0 observations)" % (self.id) 

    def _to_json(self):
        json_dataset = []
        this_ds = {}
        this_ds["id"] = self.id
        this_ds["observations"] = self.observations
        this_ds["results"] = self.results
        json_dataset.append(this_ds)
        return json_dataset

    


import os
import json

def _get_creds():
    with open(os.path.dirname(os.path.realpath(__file__))+"/.local/credentials.json") as file:
        json_credentials=json.loads(file.read())
    return json_credentials

creds = _get_creds()
username = creds['username']
password = creds['password']
agena_key = creds['agena-key']



import unittest;
import pyagena;

class TestCreateAsia(unittest.TestCase):
    def test_create_asia(self):
        model = pyagena.Model()
        model.change_settings(convergence=0.01)

        net = model.create_network(id = 'Asia?_0', name = 'Asia?', description = '.Asia?')

        nVisitAsia = net.create_node(id='A', name='Visit to Asia?',type='Labelled', states=['yes', 'no'])
        nVisitAsia.set_probabilities([[
            0.01,
            0.99
        ]])

        nHasTuberculosis = net.create_node(id='T', name='Has tuberculosis',type='Labelled', states=['yes', 'no'])
        nHasTuberculosis.add_parent(nVisitAsia)
        nHasTuberculosis.set_probabilities([
            [0.05, 0.01],
            [0.95, 0.99]
        ], True)

        nSmoker = net.create_node(id='S', name='Smoker?',type='Labelled', states=['yes', 'no'])

        nHasLungCancer = net.create_node(id='L', name='Has lung cancer',type='Labelled', states=['yes', 'no'])
        nHasLungCancer.add_parent(nSmoker)
        nHasLungCancer.set_probabilities([
            [0.1, 0.01],
            [0.9, 0.99]
        ], True)
        
        nTuberculosisOrCancer = net.create_node(id='TBoC', name='Tuberculosis or cancer',type='Labelled', states=['yes', 'no'])
        nTuberculosisOrCancer.add_parent(nHasTuberculosis)
        nTuberculosisOrCancer.add_parent(nHasLungCancer)
        nTuberculosisOrCancer.set_probabilities([
            [1.0, 1.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0]
        ], True)

        nHasBronchitis = net.create_node(id='B', name='Has bronchitis',type='Labelled', states=['yes', 'no'])
        nHasBronchitis.add_parent(nSmoker)
        nHasBronchitis.set_probabilities([
            [0.6, 0.3],
            [0.4, 0.7]
        ], True)

        nPositiveXray = net.create_node(id='X', name='Positive X-ray?',type='Labelled', states=['yes', 'no'])
        nPositiveXray.add_parent(nTuberculosisOrCancer)
        nPositiveXray.set_probabilities([
            [0.98, 0.05],
            [0.02, 0.95]
        ], True)

        nDyspnoea = net.create_node(id='D', name='Dyspnoea?',type='Labelled', states=['yes', 'no'])
        nDyspnoea.add_parent(nHasBronchitis)
        nDyspnoea.add_parent(nTuberculosisOrCancer)
        nDyspnoea.set_probabilities([
            [0.9, 0.8, 0.7, 0.1],
            [0.1, 0.2, 0.3, 0.9]
        ], True)

        ds = model.get_dataset(dataset_index=0)

        pyagena.local_api_calculate(model, [ds.id])
        result = ds.get_result(net.id, nDyspnoea.id, 'yes')
        self.assertAlmostEqual(result, 0.43597, delta=0.01)
        
        ds.enter_observation(net.id, nHasBronchitis.id, 'yes')
        ds.enter_observation(net.id, nHasTuberculosis.id, 'yes')
        pyagena.local_api_calculate(model, [ds.id])
        result = ds.get_result(net.id, nDyspnoea.id, 'yes')
        self.assertAlmostEqual(result, 0.9, delta=0.01)



import os;
import unittest;
import pyagena;

class Test2VarObservations(unittest.TestCase):
    def test_2_var_observations(self):
        model = pyagena.Model.from_cmpx(os.path.join("tests", "resources", "2obs.json"))
        ds = model.get_dataset(dataset_index=0)
        net = model.get_network(network_index=0)
        node = net.get_node(node_index=0)

        self.assertEqual(len(ds.observations), 2)
        self.assertEqual(len(ds.results), 1)

        ds.enter_observation(network_id=net.id, node_id=node.id, value=10, variable_name="a")
        ds.enter_observation(network_id=net.id, node_id=node.id, value=20, variable_name="b")

        pyagena.local_api_calculate(model, [ds.id])

        self.assertEqual(len(ds.observations), 2)
        self.assertEqual(len(ds.results), 1)

        self.assertEqual(ds.get_result(net.id, node.id).summaryStatistics.mean, 30)
        
        ds.clear_all_observations()
        self.assertEqual(len(ds.observations), 0)
        pyagena.local_api_calculate(model, [ds.id])

        self.assertEqual(ds.get_result(net.id, node.id).summaryStatistics.mean, 3)



import os;
import tempfile;
import json
import unittest
import pyagena
import tests.credentials as credentials
from tests.template_load_cars import TemplateLoadCars

class TestBatch(TemplateLoadCars):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)
        pyagena.local_api_activate_license(credentials.agena_key)

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.TemporaryDirectory()

    def test_batch(self):
        self.model.datasets = []
        self.model.import_data(os.path.join("tests", "resources", "batch_data.csv"))

        cache_file=os.path.join(self.tempdir.name, "batch_cache_dyn.json")

        pyagena.local_api_calculate(self.model, cache_path=cache_file)

        with open(cache_file) as file:
            json_cache = json.loads(file.read())
        
        self.assertEqual(len(json_cache), 100)

        file_export_inputs_json=os.path.join(self.tempdir.name, "batch_export_inputs.json")
        self.model.export_data(filename=file_export_inputs_json, include_inputs=True, include_outputs=False)
        with open(file_export_inputs_json) as file:
            jFile = json.loads(file.read())
            self.assertEqual(len(jFile), 100)
            for ds in jFile:
                self.assertTrue('id' in ds)
                self.assertTrue(len(ds['observations']) > 0)
                self.assertTrue('results' not in ds or len(ds['results']) == 0)

        file_export_outputs_json=os.path.join(self.tempdir.name, "batch_export_outputs.json")
        self.model.export_data(filename=file_export_outputs_json, include_inputs=False, include_outputs=True)
        with open(file_export_outputs_json) as file:
            jFile = json.loads(file.read())
            self.assertEqual(len(jFile), 100)
            for ds in jFile:
                self.assertTrue('id' in ds)
                self.assertTrue(len(ds['results']) > 0)
                self.assertTrue('observations' not in ds or len(ds['observations']) == 0)

        file_export_full_json=os.path.join(self.tempdir.name, "batch_export_full.json")
        self.model.export_data(filename=file_export_full_json, include_inputs=True, include_outputs=True)
        with open(file_export_full_json) as file:
            jFile = json.loads(file.read())
            self.assertEqual(len(jFile), 100)
            for ds in jFile:
                self.assertTrue('id' in ds)
                self.assertTrue(len(ds['results']) > 0)
                self.assertTrue(len(ds['observations']) > 0)


        file_export_inputs_csv=os.path.join(self.tempdir.name, "batch_export_inputs.csv")
        self.model.export_data(filename=file_export_inputs_csv, include_inputs=True, include_outputs=False)
        with open(file_export_inputs_csv) as file:
            line_count=0
            for line in file:
                if not line.isspace():
                    line_count+=1
            self.assertEqual(line_count, 349)

        file_export_outputs_csv=os.path.join(self.tempdir.name, "batch_export_outputs.csv")
        self.model.export_data(filename=file_export_outputs_csv, include_inputs=False, include_outputs=True)
        with open(file_export_outputs_csv) as file:
            line_count=0
            for line in file:
                if not line.isspace():
                    line_count+=1
                if line_count > 10000:
                    break
            self.assertTrue(line_count > 10000)
        
        file_export_outputs_csv_excel=os.path.join(self.tempdir.name, "batch_export_outputs_excel.csv")
        self.model.export_data(filename=file_export_outputs_csv_excel, include_inputs=False, include_outputs=True, excel_compatibility=True)
        with open(file_export_outputs_csv_excel) as file:
            file.readline() # Skip header line
            line=file.readline().strip('\n')
            cell_state=line.split(',')[3]
            self.assertTrue(cell_state.startswith("\"=\"\"") and cell_state.endswith("\"\"\""))

        self.model.datasets=[]
        self.assertEqual(len(self.model.datasets), 0)
        self.model.import_data(file_export_full_json)
        self.assertEqual(len(self.model.datasets), 100)

        file_full_batch_model=os.path.join(self.tempdir.name, "batch_imported_from_json.cmpx")
        self.model.save_to_file(file_full_batch_model)
        model2 = pyagena.Model.from_cmpx(file_full_batch_model)
        self.assertEqual(len(model2.datasets), 100)

    def tearDown(self):
        self.tempdir.cleanup()



from .node import Node
from .network import Network
from .dataset import Dataset, dotdict
from .model import Model

import requests as re
from getpass import getpass
import time    
import json
import logging

class login():

    access_token = None
    refresh_token = None

    def __init__(self, username = None, password = None):
        self.username = input("Enter username: ") if username is None else username
        self.password = getpass("Enter password: ") if password is None else password

        self.authenticate()

    def authenticate(self):
        login_url = "https://auth.agena.ai/realms/cloud/protocol/openid-connect/token" #auth endpoint
        login_header = {"Content-Type":"application/x-www-form-urlencoded"}
        login_body = {"client_id":"agenarisk-cloud",
                "username":self.username,
                "password":self.password,
                "grant_type":"password"}       

        login_response = re.post(login_url, headers=login_header, data=login_body)

        if login_response.status_code == 200:
            logging.info("Authentication to agena.ai cloud servers is successful")
            self.access_token = login_response.json()["access_token"]
            self.refresh_token = login_response.json()["refresh_token"]
            self.login_time = int(time.time())
            access_duration = login_response.json()["expires_in"]
            refresh_duration = login_response.json()["refresh_expires_in"]
            self.access_expire = self.login_time + access_duration
            self.refresh_expire = self.login_time + refresh_duration
            self.debug = False
            self.server = "https://api.agena.ai"

        else:
            raise ValueError("Authentication failed")

    def __repr__(self) -> str:
        return f"agena.ai cloud user ({self.username})"
    
    def set_debug(self, debug:bool):
        if debug:
            self.debug = True
            logging.info("Cloud operation results will display detailed debugging messages")
        if not debug:
            self.debug = False
            logging.info("Clod operation results will not display detailed debug messages")

    def set_server_url(self, url):
        last_char = url[-1]
        if last_char == "/":
            url = url[:-1]

        self.server = url
        logging.info(f"The root of the server URL for cloud operations is set as {url}")

    def reset_server_url(self):
        self.server = "https://api.agena.ai"
        logging.info(f"The root of the server URL for cloud operations is reset to https://api.agena.ai")

    def refresh_auth(self):
        ref_url = "https://auth.agena.ai/realms/cloud/protocol/openid-connect/token"
        ref_header = {"Content-Type":"application/x-www-form-urlencoded"}
        ref_body = {"client_id":"agenarisk-cloud",
                "refresh_token": self.refresh_token,
                "grant_type":"refresh_token"}
    
        ref_response = re.post(ref_url, headers=ref_header, data=ref_body)
        if ref_response.status_code == 200:
            self.access_token = ref_response.json()["access_token"]   

    def calculate(self, model:Model, dataset_id=None):
        now = int(time.time())
        model_to_send = model._generate_cmpx()

        calculate_url = self.server + "/public/v1/calculate"
        
        if dataset_id is None:
            calculate_body = {"sync-wait":"true", "model":model_to_send["model"]}
        else:
            for ds in model.datasets:
                if ds.id == dataset_id:
                    dataset_to_send = {"observations":ds.observations}
            calculate_body = {"sync-wait":"true", "model":model_to_send["model"], "dataSet":dataset_to_send}

        if now > self.refresh_expire:
            raise ValueError("Login has expired")
        
        if now > self.access_expire and now < self.refresh_expire:
            self.refresh_auth()

        calculate_response = re.post(calculate_url, headers={"Authorization":f"Bearer {self.access_token}"},json=calculate_body)

        if calculate_response.status_code == 200:
            logging.info(calculate_response.json()["messages"])
            if self.debug:
                for db in calculate_response.json()["debug"]:
                    logging.info(db)
            
            if calculate_response.json()["status"]=="success":
                if dataset_id is None:
                    model.datasets[0].results = calculate_response.json()["results"]
                    model.dataset[0]._convert_to_dotdict()
                else:
                    for ds in model.datasets:
                        if ds.id == dataset_id:
                            ds.results = calculate_response.json()["results"]
                            ds._convert_to_dotdict()
        elif calculate_response.status_code == 202:           
            logging.info(calculate_response.json()["messages"])
            logging.info("Polling has started, polling for calculation results will update every 3 seconds")
            
            polling_url = calculate_response.json()["pollingUrl"]
            poll_status = 202

            while poll_status == 202:
                poll_now = int(time.time())
                if poll_now > self.refresh_expire:
                    raise ValueError("Login has expired")
        
                if poll_now > self.access_expire and poll_now < self.refresh_expire:
                    self.refresh_auth()

                polled_response = re.get(polling_url, headers={"Authorization":f"Bearer {self.access_token}"})
                poll_status = polled_response.status_code
                time.sleep(3)

            if polled_response.status_code == 200:
                logging.info(polled_response.json()["messages"])
                if self.debug:
                    for db in polled_response.json()["debug"]:
                        logging.info(db)

                if polled_response.json()["status"]=="success":
                    if dataset_id is None:
                        model.datasets[0].results = polled_response.json()["results"]
                        model.datasets[0]._convert_to_dotdict()
                    else:
                        for ds in model.datasets:
                            if ds.id == dataset_id:
                                ds.results = polled_response.json()["results"]
                                ds._convert_to_dotdict()
                
            else:
                if self.debug:
                    for db in polled_response.json()["debug"]:
                        logging.info(db)
                raise ValueError(polled_response.json()["messages"]) 
        
        else:
            if self.debug:
                for db in calculate_response.json()["debug"]:
                    logging.info(db)
            raise ValueError(calculate_response.json()["messages"])
        
    def sensitivity_analysis(self, model:Model, sens_config):

        def _results_to_dotdict(input):
            dot_results = dotdict(input)
            dot_results.results = dotdict(dot_results.results)
            for idx, tb in enumerate(dot_results.results.tables):
                dot_results.results.tables[idx] = dotdict(dot_results.results.tables[idx])
            for idx, cur in enumerate(dot_results.results.responseCurveGraphs):
                dot_results.results.responseCurveGraphs[idx] = dotdict(dot_results.results.responseCurveGraphs[idx])
            for idx, tor in enumerate(dot_results.results.tornadoGraphs):
                dot_results.results.tornadoGraphs[idx] = dotdict(dot_results.results.tornadoGraphs[idx])
        
            return dot_results

        now = int(time.time())
        model_to_send = model._generate_cmpx()
        sa_url = self.server + "/public/v1/tools/sensitivity"
        
        sa_body = {"sync-wait":"true", "model":model_to_send["model"], "sensitivityConfig":sens_config}

        if now > self.refresh_expire:
            raise ValueError("Login has expired")
        
        if now > self.access_expire and now < self.refresh_expire:
            self.refresh_auth()

        sa_response = re.post(sa_url, headers={"Authorization":f"Bearer {self.access_token}"},json=sa_body)

        if sa_response.status_code == 200:
            logging.info(sa_response.json()["messages"])
            if self.debug:
                for db in sa_response.json()["debug"]:
                    logging.info(db)
            
            if sa_response.json()["status"]=="success":
                sa_results = {}
                fields = ["lastUpdated", "version", "log", "uuid", "debug", "duration", "messages", "results", "memory"]
                for f in fields:
                    sa_results[f] = sa_response.json()[f]
                sa_results = _results_to_dotdict(sa_results)
        
        elif sa_response.status_code == 202:
            logging.info(sa_response.json()["messages"])
            logging.info("Polling has started, polling for calculation results will update every 3 seconds")
            
            polling_url = sa_response.json()["pollingUrl"]
            poll_status = 202

            while poll_status == 202:
                poll_now = int(time.time())
                if poll_now > self.refresh_expire:
                    raise ValueError("Login has expired")
        
                if poll_now > self.access_expire and poll_now < self.refresh_expire:
                    self.refresh_auth()

                polled_response = re.get(polling_url, headers={"Authorization":f"Bearer {self.access_token}"})
                poll_status = polled_response.status_code
                time.sleep(3)

            if polled_response.status_code == 200:
                logging.info(polled_response.json()["messages"])
                if self.debug:
                    for db in polled_response.json()["debug"]:
                        logging.info(db)

                if polled_response.json()["status"]=="success":
                    sa_results = {}
                    fields = ["lastUpdated", "version", "log", "uuid", "debug", "duration", "messages", "results", "memory"]
                    for f in fields:
                        sa_results[f] = polled_response.json()[f]
                    sa_results = _results_to_dotdict(sa_results)
                
            else:
                if self.debug:
                    for db in polled_response.json()["debug"]:
                        logging.info(db)
                raise ValueError(polled_response.json()["messages"])
                
        else:
            if self.debug:
                for db in sa_response.json()["debug"]:
                    logging.info(db)
            raise ValueError(sa_response.json()["messages"])
        
        return sa_results
    




import tempfile;
import os;
import unittest;
import pyagena;

class TestNonStringIds(unittest.TestCase):
    def test_non_string_ids(self):
        model = pyagena.Model()
        idn1=123
        idn2=321
        model.create_network(idn1)
        model.get_network().create_node(idn2)

        tempdir = tempfile.TemporaryDirectory()
        outpath = os.path.join(tempdir.name, "test_non_string_ids.cmpx")
        model.save_to_file(outpath)

        model = pyagena.Model.from_cmpx(outpath)

        self.assertEqual(model.get_network(str(idn1)).get_node(str(idn2)).id, str(idn2))

        tempdir.cleanup()



from .node import Node

import logging
import networkx as nx
import matplotlib.pyplot as plt

class Network():
    def __init__(self, id, name=None, description=None, nodes=None):
            
        self.id = str(id)

        if name is None:
            self.name = self.id
        else:
            self.name = str(name)
        
        if description is None:
            self.description = "New Network"
        else:
            self.description = str(description)
        
        if nodes is not None:
            self.nodes = nodes
        else:
            self.nodes = []
        
    def plot(self):

        from_list = []
        to_list = []

        for nd in self.nodes:
            if len(nd.parents)>0:
                for pr in nd.parents:
                    from_list.append(pr.id)
                    to_list.append(nd.id)
        
        G = nx.DiGraph()
        for nd in self._get_nodes():
            G.add_node(nd)

        for p, c in zip(from_list, to_list):
            G.add_edges_from([(p, c)])

        nx.draw(G,with_labels=True)
        plt.draw()
        plt.show()

    def create_node(self, id, name=None, description=None, type=None, simulated=None, states=None):
        this_node = Node(id=id, name=name, description=description, type=type, simulated=simulated, states=states)
        self.add_node(this_node)
        return this_node

    def create_edge(self, child_id, parent_id):
        child_node = self.get_node(child_id)
        parent_node = self.get_node(parent_id)
        child_node.add_parent(parent_node)

    def remove_edge(self, child_id, parent_id):
        child_node = self.get_node(child_id)
        parent_node = self.get_node(parent_id)        

    def set_node_probabilities(self, node_id, probabilities, by_row=False):
        node = self.get_node(node_id)
        node.set_probabilities(probabilities=probabilities, by_row=by_row)

    def set_node_states(self, node_id, states):
        node = self.get_node(node_id)
        node.set_states(states)

    def set_node_expressions(self, node_id, expressions, partitioned_parents=None, from_cmpx=False):
        node = self.get_node(node_id)
        node.set_expressions(expressions=expressions, partitioned_parents=partitioned_parents, from_cmpx=from_cmpx)
    
    def set_node_variable(self, node_id, variable_name, variable_value, from_cmpx=False):
        node = self.get_node(node_id)
        node.set_variable(variable_name=variable_name, variable_value=variable_value, from_cmpx=from_cmpx)

    def set_node_distr_type(self, node_id, distr_type, from_cmpx=False):
        node = self.get_node(node_id)
        node.set_distr_type(distr_type=distr_type, from_cmpx=from_cmpx)

    def add_node(self, new_node: Node):
        if new_node.id in self._get_nodes():
            raise ValueError(f"There is already a node in the network with the id {new_node.id}")
        else:
            self.nodes.append(new_node)
            logging.info(f"The node {new_node.id} is successfully added to the network")

    def remove_node(self, node_id):  
        old_node = self.get_node(node_id)
        if old_node in self.nodes:
            self.nodes.remove(old_node)
            logging.info(f"The node {node_id} is successfully removed from the network - if {node_id} had any child nodes in the network, make sure to adjust their parents accordingly")
        else:
            raise ValueError(f"The network {self.id} does not have a node with the id {node_id}")

    def _get_nodes(self):
        nodes_list = []
        if len(self.nodes)>0:
            for nd in self.nodes:
                nodes_list.append(nd.id)
            
        return nodes_list
    
    def get_node(self, node_id=None, node_index = 0) -> Node:
          
        if node_id is None:
            node = self.nodes[node_index]
        
        else:
            if node_id not in self._get_nodes():
                raise ValueError(f"The network {self.id} does not have a node with the id {node_id}")
          
            node = [n for n in self.nodes if n.id==node_id].pop()
        
        return node
    
    def __str__(self) -> str:
        if self.nodes is not None:
            return  "Network id: % s\nNetwork name: % s\nNetwork nodes: % s" % (self.id, self.name, ", ".join(self._get_nodes()))
        else:
            return  "Network id: % s\nNetwork name: % s" % (self.id, self.name)        

    def __repr__(self) -> str:
        return "Bayesian Network: % s (% s)" % (self.name, ", ".join(self._get_nodes())) 



import setuptools

setuptools.setup(
    name="pyagena",
    version="1.0.0",
    author="Erhan Pisirir, Eugene Dementiev",
    author_email="support@agenarisk.com",
    description="Python wrapper for agena.ai to create Bayesian network models from scratch or import existing models and export to agena.ai cloud or local API for calculations.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/AgenaRisk/api-py",
    download_url="https://github.com/AgenaRisk/api-py/archive/refs/tags/1.0.0.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['requests','pandas','networkx','matplotlib'],
    include_package_data=True
)


import unittest;
import pyagena;

class TestModelSettings(unittest.TestCase):
    def test_model_settings(self):
        model = pyagena.Model()
        model.create_network("net")
        model.get_network().create_node("node")

        # print(model.settings)

        self.assertEqual(model.settings['parameterLearningLogging'], False)
        self.assertEqual(model.settings['discreteTails'], False)
        self.assertEqual(model.settings['sampleSizeRanked'], 5)
        self.assertEqual(model.settings['convergence'], 0.01)
        self.assertEqual(model.settings['simulationLogging'], False)
        self.assertEqual(model.settings['iterations'], 50)
        self.assertEqual(model.settings['tolerance'], 1)

        model.change_settings(iterations=30, convergence=0.001)
        self.assertEqual(model.settings['convergence'], 0.001)
        self.assertEqual(model.settings['iterations'], 30)

        model.change_settings(iterations=45)
        self.assertEqual(model.settings['convergence'], 0.001)
        self.assertEqual(model.settings['iterations'], 45)
        


from .node import Node
from .network import Network
from .dataset import Dataset

import json
import pandas as pd
import os
import logging

class Model():         
     def __init__(self, networks=None, id=None, datasets=None, network_links=None, settings=None):
          
          if networks is not None:
               self.networks = networks
          else:
               self.networks = []

          if settings is None:
               self.settings = {"parameterLearningLogging":False, "discreteTails":False, "sampleSizeRanked":5, "convergence":0.01, "simulationLogging":False, "iterations":50, "tolerance":1}
          else:
               self.settings = settings

          if id is not None:
               self.id = str(id)
          else:
               self.id = None

          if datasets is None:
               self.datasets = []
               self.datasets.append(Dataset(id="Case 1"))
          else:
               self.datasets = datasets

          if network_links is None:
               self.network_links = []
          else:
               self.network_links = network_links
          
     def __str__(self) -> str:
          if self.id is not None:
               return  "Model id: % s\nModel networks: % s" % (self.id, ", ".join(self._get_networks()))   
          else:
               return  "Model networks: % s" % (", ".join(self._get_networks()))   
     
     def __repr__(self) -> str:
          if self.id is not None:
               return "Bayesian Model: % s (% s)" % (self.id, ", ".join(self._get_networks())) 
          else:
               return "Bayesian Model: (% s)" % (", ".join(self._get_networks())) 

     def create_network(self, id, name=None, description=None):
          this_net = Network(id=id, name=name, description=description)
          self.add_network(this_net)
          return this_net

     def add_network(self, network:Network):
          if network.id in self._get_networks():
               raise ValueError("There is already a network with this id in the model")
          else:
               self.networks.append(network)
               logging.info(f"The network {network.id} is successfully added to the model")

     def remove_network(self, id):
          network = self.get_network(id)
          if network not in self.networks:
               raise ValueError("This network is not in the model")
          else:
               self.networks.remove(network)
               logging.warning(f"The network is successfully removed from the model - if {network.id} had any links to other networks in the model, make sure to adjust network links accordingly")

     def _get_networks(self):
        nets_list = []
        if len(self.networks)>0:
            for nt in self.networks:
                nets_list.append(nt.id)
            
        return nets_list

     def get_network(self, network_id=None, network_index = 0) -> Network:
          if network_id is None:
               network = self.networks[network_index]
          else:
               if network_id not in self._get_networks():
                    raise ValueError(f"The model does not have a network with the id {network_id}")
          
               network = [n for n in self.networks if n.id==network_id].pop()
          
          return network

     def get_dataset(self, dataset_id=None, dataset_index = 0) -> Dataset:
          if dataset_id is None:
               dataset = self.datasets[dataset_index]
          else:
               if dataset_id not in self._get_datasets():
                    raise ValueError(f"The model does not have a dataset with the id {dataset_id}")
              
               dataset = [d for d in self.datasets if d.id==dataset_id].pop()
          
          return dataset

     def add_network_link(self, source_network_id, source_node_id, target_network_id, target_node_id, link_type="Marginals", pass_state=None):
          out_nw = self.get_network(source_network_id)
          out_node = out_nw.get_node(source_node_id)
          in_nw = self.get_network(target_network_id)
          in_node = in_nw.get_node(target_node_id)

          if len(in_node.parents)>0:
               raise ValueError("The target node cannot be a node with parents in its network")
          
          #if in_node already a target_node raise error

          targets = []
          if len(self.network_links)>0:
               for nl in self.network_links:
                    targets.append(nl["targetNode"])
          
          if target_node_id in targets:
               raise ValueError("The required target node is already a target for an existing network link")

          num_list = ["ContinuousInterval", "IntegerInterval", "DiscreteReal"]
          num_intv_list = ["ContinuousInterval", "IntegerInterval"]
          valid = True

          if link_type == "Marginals":
               if not out_node.simulated and not in_node.simulated:
                    pass
               else:
                    if out_node.type in num_list and in_node.type in num_list:
                         pass
                    else:
                         valid = False

          if link_type in ["Mean", "Median", "Variance", "StandardDeviation", "LowerPercentile", "UpperPercentile"]:
               if out_node.type in num_list & in_node.simulated:
                    pass
               else:
                    valid = False

          if link_type == "State":
               if out_node.type not in num_intv_list and in_node.simulated:
                    pass
               else:
                    valid = False

          if valid:
               if link_type == "State":
                    if pass_state is None:
                         raise ValueError("Please enter the source node state to be passed on")
                    else:
                         new_link = {"sourceNode":source_node_id,
                                   "sourceNetwork":source_network_id,
                                   "targetNode":target_node_id,
                                   "targetNetwork":target_network_id,
                                   "passState":pass_state,
                                   "type":link_type}
               else:
                    new_link = {"sourceNode":source_node_id,
                                   "sourceNetwork":source_network_id,
                                   "targetNode":target_node_id,
                                   "targetNetwork":target_network_id,
                                   "type":link_type}
               
               self.network_links.append(new_link)
          else:
               raise ValueError("The link between source node and target node is not valid")                    

     def remove_network_link(self, source_network_id, source_node_id, target_network_id, target_node_id):

          if len(self.network_links)==0:
               raise ValueError("This model has no network links")
          
          link_to_remove = [nl for nl in self.network_links if nl["sourceNode"]==source_node_id and nl["sourceNetwork"]==source_network_id and nl["targetNode"]==target_node_id and nl["targetNetwork"]==target_network_id].pop()
          
          if link_to_remove in self.network_links:
               self.network_links.remove(link_to_remove)
               logging.info("The network link is successfully removed from the model")
          else:
               raise ValueError("This network link does not exist in the model")

     def remove_all_network_links(self):
          self.network_links = []
          logging.info("All network links are removed from the model")

     def create_dataset(self, dataset_id, observations = None, from_cmpx=False):
          new_ds = Dataset(id=dataset_id, observations=observations)
          self.datasets.append(new_ds)
          if not from_cmpx:
               logging.info(f"The dataset {dataset_id} is successfully created")
          return new_ds

     def remove_dataset(self, dataset_id, from_cmpx=False):
          if dataset_id not in self._get_datasets():
               raise ValueError(f"The dataset {dataset_id} does not exist in the model")

          del_ds = self.get_dataset(dataset_id)
          self.datasets.remove(del_ds)
          if not from_cmpx:
               logging.info(f"The dataset {dataset_id} is successfully removed from the model")

     def _get_datasets(self):
          ds_list = []

          if len(self.datasets)>0:
               for ds in self.datasets:
                    ds_list.append(ds.id)
          
          return ds_list
     
     def enter_observation(self, network_id, node_id, value, dataset_id=None, variable_name=None):
          if dataset_id is None:
               ds = self.datasets[0]
          else:
               if dataset_id not in self._get_datasets():
                    raise ValueError(f"The dataset {dataset_id} does not exist")
               else:
                    ds = self.get_dataset(dataset_id)
          
          ds.enter_observation(network_id=network_id, node_id=node_id, value=value, variable_name=variable_name)
                    
     def remove_observation(self, network_id, node_id, dataset_id=None):
          if dataset_id is None:
               ds = self.datasets[0]
          else:
               if dataset_id not in self._get_datasets():
                    raise ValueError(f"The dataset {dataset_id} does not exist")
               
               ds = self.get_dataset(dataset_id)

          ds.remove_observation(network_id=network_id, node_id=node_id)

     def clear_dataset_observations(self, dataset_id):
          if dataset_id not in self._get_datasets():
               raise ValueError(f"The dataset {dataset_id} does not exist")

          ds = self.get_dataset(dataset_id)     

          ds.clear_all_observations()

     def clear_all_observations(self):
          for ds in self.datasets:
               ds.observations = []
          
          logging.info("All observations in all datasets in the model are successfully cleared")

     def change_settings(self, **kwargs):
          for fld in kwargs:
               if fld in self.settings.keys():
                    self.settings[fld] = kwargs[fld]
                    logging.info(f"Model setting {fld} is successfully updated")

     def default_settings(self):
          self.settings = {"parameterLearningLogging":False, "discreteTails":False, "sampleSizeRanked":5, "convergence":0.01, "simulationLogging":False, "iterations":50, "tolerance":1}
          logging.info("Model settings are successfully reset to default")

     def save_to_file(self, filename, strip_data = False):
          if os.path.splitext(filename)[1] == ".cmpx" or os.path.splitext(filename)[1] == ".json":
               exported = self._generate_cmpx(strip_data=strip_data)
               
               with open(filename, "w") as outfile:
                    json.dump(exported, outfile)
          else:
               raise ValueError("The input filename must have a .cmpx or .json file extension")

     def _generate_cmpx(self, strip_data = False):
          json_settings = self.settings
          json_datasets = []

          if not strip_data:
               for ds in self.datasets:
                    this_ds = {}
                    this_ds["id"] = ds.id
                    this_ds["observations"] = ds.observations
                    this_ds["results"] = ds.results
                    json_datasets.append(this_ds)

          json_network_links = self.network_links

          json_networks = []
          for nt in self.networks:
               this_nt = {}
               this_nt["id"] = nt.id
               this_nt["name"] = nt.name
               this_nt["description"] = nt.description
               this_nt["nodes"] = []
               this_nt["links"] = []
               
               for nd in nt.nodes:
                    this_nd = {}
                    this_nd["id"] = nd.id
                    this_nd["name"] = nd.name
                    this_nd["description"] = nd.description
                    this_nd["configuration"] = {}
                    if nd.simulated:    this_nd["configuration"]["simulated"] = True
                    this_nd["configuration"]["type"] = nd.type
                    if nd.states is not None:   this_nd["configuration"]["states"] = nd.states
                    this_nd["configuration"]["table"] = {}
                    this_nd["configuration"]["table"]["type"] = nd.distr_type
                    if nd.probabilities is not None:    this_nd["configuration"]["table"]["probabilities"] = list(map(list, zip(*nd.probabilities)))
                    if nd.expressions is not None:  this_nd["configuration"]["table"]["expressions"] = nd.expressions
                    if nd.distr_type == "Partitioned" and nd.partitions is not None:   this_nd["configuration"]["table"]["partitions"] = nd.partitions
                    
                    if len(nd.variables)>0:
                         this_variables = nd.variables
                         this_nd["configuration"]["variables"] = []
                         for vr in this_variables:
                              this_nd["configuration"]["variables"].append({"name":[*vr].pop(),"value":[*vr.values()].pop()})
                    
                    this_nt["nodes"].append(this_nd)

                    this_parents = nd.parents
                    for pr in this_parents:
                         this_nt["links"].append({"parent":pr.id,"child":nd.id})
        
               json_networks.append(this_nt)

          json_model = {"model":{"settings":json_settings, "dataSets":json_datasets, "links": json_network_links, "networks": json_networks}}
          return json.loads(json.dumps(json_model))

     def import_data(self, filename):
          if os.path.splitext(filename)[1] == ".csv":
               data = pd.read_csv(filename)
               data = data.astype({data.columns[0]:"str"})
               data = data.set_index(data.columns[0])

               for ix, row in data.iterrows():
                    ds_id = ix
                    if ds_id in self._get_datasets():
                         self.remove_dataset(ds_id, from_cmpx=True)
                    self.create_dataset(ix, from_cmpx=True)
                    logging.info(f"The dataset {ds_id} is created or updated in the model")
                    for idx, cl in enumerate(data.columns):
                         this_net = cl.split(".")[0]
                         this_node = cl.split(".")[1]
                         if not pd.isna(row[idx]):     self.enter_observation(network_id=this_net, node_id=this_node, dataset_id=ds_id, value=row[idx])
                    
    
          elif os.path.splitext(filename)[1] == ".json":
               with open(filename, "r") as file:
                    datasets_string = file.read()
               new_datasets = json.loads(datasets_string)

               for ds in new_datasets:
                    ds_id = ds["id"]
                    if ds_id in self._get_datasets():
                         self.remove_dataset(ds_id, from_cmpx=True)
                    self.create_dataset(dataset_id = ds_id, observations=ds["observations"], from_cmpx=True)
                    
                    if "results" in ds.keys():
                         this_ds = self.get_dataset(ds_id)
                         this_ds.results = ds["results"]                   
          
                    logging.info(f"The dataset {ds_id} is created or updated in the model")
                                   
          else:
               raise ValueError("Data file should be in .csv or .json format")
          
     def create_csv_template(self, filename=None):
          headers = [(nt.id+"."+nd.id) for nt in self.networks for nd in nt.nodes]
          rows = [ds.id for ds in self.datasets]
          test = pd.DataFrame(columns=headers, index=rows)
          test.index.name = "Case"
          if filename is None:
               filename = "New_Data_Template.csv"
          test.to_csv(filename)

     def _import_results(self, file_path):
          with open(file_path, "r") as file:
               results_string = file.read()
          all_results = json.loads(results_string)
          for res in all_results:
               results = res["results"]
               dataset_id = res["id"]
               dataset = self.get_dataset(dataset_id)
               dataset.results = results
               dataset._convert_to_dotdict()
               logging.info(f"Results are successfully imported to case {dataset.id}")

     def export_data(self, filename, dataset_ids=None, include_inputs = False, include_outputs = True, excel_compatibility = False):

          if dataset_ids is None:
               ds_to_export = self.datasets
          else:
               ds_to_export = [ds for ds in self.datasets if ds.id in dataset_ids]

          if os.path.splitext(filename)[1] == ".csv":
               
               if include_inputs & include_outputs:
                    raise ValueError("For the csv file please choose only either inputs or outputs to export")
               
               if include_outputs:
                    df = pd.DataFrame(columns=["Case", "Network", "Node", "State", "Probability"])

                    for ds in ds_to_export:
                         for rs in ds.results:
                              for rv in rs.resultValues:
                                   if excel_compatibility:
                                        df.loc[len(df)] = [ds.id, rs["network"], rs["node"], '="'+rv["label"]+'"', rv["value"]]
                                   else:
                                        df.loc[len(df)] = [ds.id, rs["network"], rs["node"], rv["label"], rv["value"]]
                                       
                    df.to_csv(filename, index=False)

               if include_inputs:
                    df = pd.DataFrame(columns=["Case", "Network", "Node", "State", "Probability"])

                    for ds in ds_to_export:
                         for obs in ds.observations:
                              for ent in obs.entries:
                                   df.loc[len(df)] = [ds.id, obs.network, obs.node, ent.value, ent.weight/len(obs.entries)]
                    
                    df.to_csv(filename, index=False)

          elif os.path.splitext(filename)[1] == ".json":

               if dataset_ids is not None:
                    data_json = self._ds_to_json(dataset_ids)
               else:
                    data_json = self._ds_to_json()
               
               if not include_inputs:
                    for ds in data_json:
                         ds["observations"] = []
               
               if not include_outputs:
                    for ds in data_json:
                         ds["results"] = []

               with open(filename, "w") as outfile:
                    json.dump(data_json, outfile)

          else:
               raise ValueError("The exported file should be in either .csv or .json format")  

     def get_result(self, network_id, node_id, node_state = None, dataset_id = None):
          if dataset_id is None:
               ds = self.datasets[0]
          else:
               if dataset_id not in self._get_datasets():
                    raise ValueError(f"The dataset {dataset_id} does not exist")
               else:
                    ds = self.get_dataset(dataset_id)
               
          result = ds.get_result(network_id=network_id, node_id=node_id, node_state=node_state)
          return result

     def export_node_results(self, network_id, node_id, filename=None):
          this_net = self.get_network(network_id)
          this_node = this_net.get_node(node_id=node_id)
          rows = self._get_datasets()
          columns = this_node.states
          output_results = pd.DataFrame(columns=columns, index=rows)

          for dt in self.datasets:
               res = dt.get_result(network_id=network_id, node_id=node_id)
               res_values = res.resultValues
               for vl in res_values:
                    for st in this_node.states:
                         if vl.label == st:
                              output_results.loc[dt.id,st] = vl.value

          output_results.index.name = "Case"
          if filename is None:
               filename = node_id + "_batch_results.csv"
          output_results.to_csv(filename)

     def create_sensitivity_config(self, **kwargs):
          sens_config = {}
          for fld in kwargs:
               sens_config[fld] = kwargs[fld]
          return sens_config

     def _ds_to_json(self, dataset_ids = None):
          json_dataset = []
          if dataset_ids is None:
               for ds in self.datasets:
                    this_ds = {}
                    this_ds["id"] = ds.id
                    this_ds["observations"] = ds.observations
                    this_ds["results"] = ds.results
                    json_dataset.append(this_ds)
          else:
               for dsid in dataset_ids:
                    ds = self.get_dataset(dsid)
                    this_ds = {}
                    this_ds["id"] = ds.id
                    this_ds["observations"] = ds.observations
                    this_ds["results"] = ds.results
                    json_dataset.append(this_ds)
          
          return json_dataset

     @classmethod
     def from_cmpx(cls, filename):
          
          with open(filename, "r") as file:
               model_string = file.read()

          agena_model = json.loads(model_string)["model"]
          agena_settings = agena_model["settings"]
          agena_networks = agena_model["networks"]
          agena_datasets = agena_model["dataSets"]
          agena_network_links = agena_model["links"]
          
          agena_nodes = []
          for index, anw in enumerate(agena_networks):
               agena_nodes.append([])
               for an in anw["nodes"]:
                    agena_nodes[index].append(an)

          agena_node_links = []
          for index, anw in enumerate(agena_networks):
               agena_node_links.append([])
               for al in anw["links"]:
                    agena_node_links[index].append([al["child"], al["parent"]])

          nodes_list = []
          # creating a list of nodes with id, name, description, and type
          for idx, ntw in enumerate(agena_nodes):
               nodes_list.append([])
               for nd in ntw:
                    if "simulated" not in nd["configuration"].keys():
                         this_node = Node(id=nd["id"], name=nd["name"], type=nd["configuration"]["type"], simulated=False)
                    else:
                         this_node = Node(id=nd["id"], name=nd["name"], type=nd["configuration"]["type"], simulated=True)
                    nodes_list[idx].append(this_node)

          #setting distr type and states for Manual distr types and non simulated interval nodes
          for idx, ntw in enumerate(agena_nodes):
               for ix, nd in enumerate(ntw):
                    this_node = nodes_list[idx][ix]

                    if nd["configuration"]["table"]["type"] == "Manual":
                         this_node.set_distr_type(nd["configuration"]["table"]["type"], from_cmpx=True)    

                    if "simulated" not in nd["configuration"].keys():
                         if nd["configuration"]["type"] == "ContinuousInterval" or nd["configuration"]["type"] == "IntegerInterval":
                              this_node.set_states(nd["configuration"]["states"], from_cmpx=True)                    

                    if this_node.distr_type == "Manual":
                         this_node.set_states(nd["configuration"]["states"], from_cmpx=True)                

          # adding parents to all nodes before defining probabilities, expressions, and partitions
          for idx, ntw in enumerate(nodes_list):
               for nd in ntw:
                    for link in agena_node_links[idx]:
                         if nd.id == link[0]:
                              nd._addparentbyID(ntw,link[1])

          # setting distr_type, probabilities, expressions, and partitions for all nodes
          for idx, ntw in enumerate(agena_nodes):
               for ix, nd in enumerate(ntw):
                    this_node = nodes_list[idx][ix]
                    this_node.set_distr_type(nd["configuration"]["table"]["type"], from_cmpx=True)  

                    if this_node.distr_type == "Manual":
                         this_node.set_probabilities(nd["configuration"]["table"]["probabilities"], by_row=True, from_cmpx=True)
                    if this_node.distr_type == "Expression":
                         if (this_node.states is not None) and ("probabilities" in nd["configuration"]["table"].keys()):
                              this_node.set_probabilities(nd["configuration"]["table"]["probabilities"], by_row=True, from_cmpx=True)
                         this_node.set_expressions(nd["configuration"]["table"]["expressions"], from_cmpx=True)
                    if this_node.distr_type == "Partitioned":
                         if (this_node.states is not None) and ("probabilities" in nd["configuration"]["table"].keys()):
                              this_node.set_probabilities(nd["configuration"]["table"]["probabilities"], by_row=True, from_cmpx=True)
                         this_node.set_expressions(nd["configuration"]["table"]["expressions"], partitioned_parents= nd["configuration"]["table"]["partitions"], from_cmpx=True)

                    if "variables" in nd["configuration"].keys():
                         for vr in nd["configuration"]["variables"]:
                              this_node.set_variable(vr["name"], vr["value"], from_cmpx=True)
          
          # creating a list of networks
          networks_list = []
          for idx, ntw in enumerate(agena_networks):
               this_net = Network(id=ntw["id"], nodes=nodes_list[idx])
               networks_list.append(this_net)

          datasets_list = []
          if len(agena_datasets)>0:
               for ds in agena_datasets:
                    this_ds = Dataset(id=ds["id"], observations=ds["observations"], results=ds["results"])
                    datasets_list.append(this_ds)


          return cls(networks=networks_list, id=None, datasets=datasets_list, network_links=agena_network_links, settings=agena_settings)


from tests.template_load_cars import TemplateLoadCars
import unittest
import pyagena
import tests.credentials as credentials

class TestLoadAndCalculateLocal(TemplateLoadCars):

    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)
        pyagena.local_api_activate_license(credentials.agena_key)

    def test_calculate_short(self):
        resultInitialFile = self.ds.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean
        pyagena.local_api_calculate(self.model, [self.ds.id])
        resultInitialCloud = self.ds.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean
        self.assertAlmostEqual(resultInitialCloud, 2290, delta=0.1)
        self.assertAlmostEqual(resultInitialCloud, resultInitialFile, delta=10)

        self.ds.enter_observation(network_id=self.network.id, node_id=self.nodeCarType.id, value="Small")
        self.ds.enter_observation(network_id=self.network.id, node_id=self.nodeMilesPerYear.id, value=1000)
        self.ds.enter_observation(network_id=self.network.id, node_id=self.nodeReliability.id, value="Low")
        pyagena.local_api_calculate(self.model, [self.ds.id])
        resultCustom = self.ds.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean
        self.assertAlmostEqual(resultCustom, 261, delta=10)

        self.model.create_dataset('empty')
        dsEmpty = self.model.get_dataset('empty')
        pyagena.local_api_calculate(self.model, [dsEmpty.id])
        resultEmptyDs = dsEmpty.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean
        self.assertAlmostEqual(resultEmptyDs, 1562, delta=10)

    def test_calculate_long(self):
        self.model.change_settings(convergence=0.00001)
        pyagena.local_api_calculate(self.model, [self.ds.id])
        result = self.ds.get_result(self.network.id, self.nodeTotalAnnualCost.id).summaryStatistics.mean
        self.assertAlmostEqual(result, 2269, delta=10)




import unittest;
import pyagena;

class TestCreateRiverFlooding(unittest.TestCase):
    
    def test_create_river_flooding(self):
        model = pyagena.Model()
        model.change_settings(convergence=0.01)

        ### NET_1 ###
        net_1 = model.create_network(id = 'net_1')

        nodeRain_1 = net_1.create_node(id='nodeRain_1', type='Ranked')
        nodeRain_1.states=['None', 'Low Rainfall', 'High Rainfall']
        nodeRain_1.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodePriorWaterLevel_1 = net_1.create_node(id='nodePriorWaterLevel_1', type='Ranked')
        nodePriorWaterLevel_1.set_states(['Low', 'Medium', 'High'])
        nodePriorWaterLevel_1.set_probabilities([[
            0.5,
            0.33333334,
            0.16666667
        ]])

        nodePostWaterLevel_1 = net_1.create_node(id='nodePostWaterLevel_1', type='Ranked')
        nodePostWaterLevel_1.set_states(['Low', 'Medium', 'High'])
        nodePostWaterLevel_1.add_parent(nodePriorWaterLevel_1)
        nodePostWaterLevel_1.add_parent(nodeRain_1)
        nodePostWaterLevel_1.set_probabilities([
            [1.0, 0.9, 0.2, 0.7, 0.1, 0.0, 0.1, 0.0, 0.0],
            [0.0, 0.1, 0.7, 0.3, 0.8, 0.2, 0.7, 0.1, 0.0],
            [0.0, 0.0, 0.1, 0.0, 0.1, 0.8, 0.2, 0.9, 1.0]
        ], True)

        nodeFloodDefences_1 = net_1.create_node(id='nodeFloodDefences_1', type='Ranked')
        nodeFloodDefences_1.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefences_1.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodeFloodDefencesOutput_1 = net_1.create_node(id='nodeFloodDefencesOutput_1', type='Ranked')
        nodeFloodDefencesOutput_1.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefencesOutput_1.add_parent(nodeFloodDefences_1)
        nodeFloodDefencesOutput_1.set_probabilities([
            [1.0, 0.1, 0.0],
            [0.0, 0.9, 0.1],
            [0.0, 0.0, 0.9]
        ], True)

        nodeFlood_1 = net_1.create_node(id='nodeFlood_1', type='Ranked')
        nodeFlood_1.set_states(['No', 'Yes'])
        nodeFlood_1.add_parent(nodeFloodDefences_1)
        nodeFlood_1.add_parent(nodePostWaterLevel_1)
        nodeFlood_1.set_distr_type('Expression')
        nodeFlood_1.set_expressions(['TNormal(nodePostWaterLevel_1-nodeFloodDefences_1, 0.1, 0, 1)'])

        ### NET_2 ###
        net_2 = model.create_network(id = 'net_2')

        nodeRain_2 = net_2.create_node(id='nodeRain_2', type='Ranked')
        nodeRain_2.set_states(['None', 'Low Rainfall', 'High Rainfall'])
        nodeRain_2.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodePriorWaterLevel_2 = net_2.create_node(id='nodePriorWaterLevel_2', type='Ranked')
        nodePriorWaterLevel_2.set_states(['Low', 'Medium', 'High'])
        nodePriorWaterLevel_2.set_probabilities([[
            0.5,
            0.33333334,
            0.16666667
        ]])

        nodePostWaterLevel_2 = net_2.create_node(id='nodePostWaterLevel_2', type='Ranked')
        nodePostWaterLevel_2.set_states(['Low', 'Medium', 'High'])
        nodePostWaterLevel_2.add_parent(nodePriorWaterLevel_2)
        nodePostWaterLevel_2.add_parent(nodeRain_2)
        nodePostWaterLevel_2.set_probabilities([
            [1.0, 0.9, 0.2, 0.7, 0.1, 0.0, 0.1, 0.0, 0.0],
            [0.0, 0.1, 0.7, 0.3, 0.8, 0.2, 0.7, 0.1, 0.0],
            [0.0, 0.0, 0.1, 0.0, 0.1, 0.8, 0.2, 0.9, 1.0]
        ], True)

        nodeFloodDefences_2 = net_2.create_node(id='nodeFloodDefences_2', type='Ranked')
        nodeFloodDefences_2.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefences_2.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodeFloodDefencesOutput_2 = net_2.create_node(id='nodeFloodDefencesOutput_2', type='Ranked')
        nodeFloodDefencesOutput_2.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefencesOutput_2.add_parent(nodeFloodDefences_2)
        nodeFloodDefencesOutput_2.set_probabilities([
            [1.0, 0.1, 0.0],
            [0.0, 0.9, 0.1],
            [0.0, 0.0, 0.9]
        ], True)

        nodeFlood_2 = net_2.create_node(id='nodeFlood_2', type='Ranked')
        nodeFlood_2.set_states(['No', 'Yes'])
        nodeFlood_2.add_parent(nodeFloodDefences_2)
        nodeFlood_2.add_parent(nodePostWaterLevel_2)
        nodeFlood_2.set_distr_type('Expression')
        nodeFlood_2.set_expressions(['TNormal(nodePostWaterLevel_2-nodeFloodDefences_2, 0.1, 0, 1)'])

        ### NET_3 ###
        net_3 = model.create_network(id = 'net_3')

        nodeRain_3 = net_3.create_node(id='nodeRain_3', type='Ranked')
        nodeRain_3.set_states(['None', 'Low Rainfall', 'High Rainfall'])
        nodeRain_3.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodePriorWaterLevel_3 = net_3.create_node(id='nodePriorWaterLevel_3', type='Ranked')
        nodePriorWaterLevel_3.set_states(['Low', 'Medium', 'High'])
        nodePriorWaterLevel_3.set_probabilities([[
            0.5,
            0.33333334,
            0.16666667
        ]])

        nodePostWaterLevel_3 = net_3.create_node(id='nodePostWaterLevel_3', type='Ranked')
        nodePostWaterLevel_3.set_states(['Low', 'Medium', 'High'])
        nodePostWaterLevel_3.add_parent(nodePriorWaterLevel_3)
        nodePostWaterLevel_3.add_parent(nodeRain_3)
        nodePostWaterLevel_3.set_probabilities([
            [1.0, 0.9, 0.2, 0.7, 0.1, 0.0, 0.1, 0.0, 0.0],
            [0.0, 0.1, 0.7, 0.3, 0.8, 0.2, 0.7, 0.1, 0.0],
            [0.0, 0.0, 0.1, 0.0, 0.1, 0.8, 0.2, 0.9, 1.0]
        ], True)

        nodeFloodDefences_3 = net_3.create_node(id='nodeFloodDefences_3', type='Ranked')
        nodeFloodDefences_3.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefences_3.set_probabilities([[
            0.16666667,
            0.33333334,
            0.5
        ]])

        nodeFloodDefencesOutput_3 = net_3.create_node(id='nodeFloodDefencesOutput_3', type='Ranked')
        nodeFloodDefencesOutput_3.set_states(['Poor', 'Medium', 'High'])
        nodeFloodDefencesOutput_3.add_parent(nodeFloodDefences_3)
        nodeFloodDefencesOutput_3.set_probabilities([
            [1.0, 0.1, 0.0],
            [0.0, 0.9, 0.1],
            [0.0, 0.0, 0.9]
        ], True)

        nodeFlood_3 = net_3.create_node(id='nodeFlood_3', type='Ranked')
        nodeFlood_3.set_states(['No', 'Yes'])
        nodeFlood_3.add_parent(nodeFloodDefences_3)
        nodeFlood_3.add_parent(nodePostWaterLevel_3)
        nodeFlood_3.set_distr_type('Expression')
        nodeFlood_3.set_expressions(['TNormal(nodePostWaterLevel_3-nodeFloodDefences_3, 0.1, 0, 1)'])

        model.add_network_link(
            source_network_id=net_1.id,
            source_node_id=nodePostWaterLevel_1.id,
            target_network_id=net_2.id,
            target_node_id=nodePriorWaterLevel_2.id,
            link_type="Marginals"
        )
        model.add_network_link(
            source_network_id=net_1.id,
            source_node_id=nodeFloodDefencesOutput_1.id,
            target_network_id=net_2.id,
            target_node_id=nodeFloodDefences_2.id,
        )
        model.add_network_link(
            source_network_id=net_2.id,
            source_node_id=nodePostWaterLevel_2.id,
            target_network_id=net_3.id,
            target_node_id=nodePriorWaterLevel_3.id,
        )
        model.add_network_link(
            source_network_id=net_2.id,
            source_node_id=nodeFloodDefencesOutput_2.id,
            target_network_id=net_3.id,
            target_node_id=nodeFloodDefences_3.id,
        )

        ds = model.get_dataset(dataset_index=0)

        pyagena.local_api_calculate(model, [ds.id])
        result = ds.get_result(net_3.id, nodeFlood_3.id, 'Yes')
        self.assertAlmostEqual(result, 0.22047, delta=0.01)
        
        ds.enter_observation(net_1.id, nodeRain_1.id, 'High Rainfall')
        ds.enter_observation(net_1.id, nodeFloodDefences_1.id, 'High')
        ds.enter_observation(net_2.id, nodeRain_2.id, 'Low Rainfall')
        ds.enter_observation(net_2.id, nodeFloodDefences_2.id, 'Medium')
        ds.enter_observation(net_3.id, nodeRain_3.id, 'None')
        ds.enter_observation(net_3.id, nodeFloodDefences_3.id, 'Poor')

        model.save_to_file('/x/river_flooding.cmpx')

        pyagena.local_api_calculate(model, [ds.id], verbose=True)
        result = ds.get_result(net_3.id, nodeFlood_3.id, 'Yes')
        self.assertAlmostEqual(result, 0.27927, delta=0.01)



import os;
import unittest
import pyagena

class TemplateLoadCars(unittest.TestCase):
    def setUp(self):
        self.model = pyagena.Model.from_cmpx(os.path.join("tests", "resources", "car_costs.json"))
        self.ds = self.model.get_dataset('mercedes')
        self.network = self.model.get_network(network_index=0)
        self.nodeCarType = self.network.get_node('car_type')
        self.nodeMilesPerYear = self.network.get_node('Miles_per_year')
        self.nodeReliability = self.network.get_node('Reliability')
        self.nodeTotalAnnualCost = self.network.get_node('total_cost')
